import logging

from flask import Blueprint, url_for, request, redirect, session, make_response
from flask_oauthlib.client import OAuth, OAuthRemoteApp

from portal import oauth
from portal.database.models import User
from portal.service import general

log = logging.getLogger(__name__)


def extract_user_info(me: dict) -> dict:
    return dict(
        uco=None,  # TODO: Need from gitlab or prompt the user
        name=me['name'],
        username=me['username'],
        email=me['email']
    )


def create_gitlab_app(oauth: OAuth) -> OAuthRemoteApp:
    base_url = oauth.app.config['GITLAB_URL']
    client_id = oauth.app.config['GITLAB_CLIENT_ID']
    client_secret = oauth.app.config['GITLAB_CLIENT_SECRET']

    return oauth.remote_app(
        'gitlab',
        base_url=base_url,
        request_token_url=None,
        access_token_url=base_url + "/oauth/token",
        authorize_url=base_url + "/oauth/authorize",
        access_token_method='POST',
        consumer_key=client_id,
        consumer_secret=client_secret
    )


gitlab = create_gitlab_app(oauth=oauth)
oauth_blueprint = Blueprint('oauth', __name__, url_prefix='/oauth')


@oauth_blueprint.route('/login', methods=['GET'])
def oauth_login():
    callback = url_for('oauth.oauth_authorized', _external=True, _scheme='https')
    log.debug(f"Callback set: {callback}")
    return gitlab.authorize(callback=callback)


def user_oauth_register(user_info):
    new_user = User(
        uco=None,
        email=user_info['email'],
        username=user_info['username'],
        is_admin=False
    )
    new_user.name = user_info['name']
    general.write_entity(new_user)
    log.debug(f"Created user={new_user}")
    return new_user


def user_login(user_info):
    user = general.find_user(user_info['username'], throws=False)
    if not user:
        user = user_oauth_register(user_info)
    resp = make_response(redirect(oauth.app.config.get('FRONTEND_URL')))
    resp.set_cookie('username', user.username)
    return resp


@oauth_blueprint.route('/login/authorized', methods=['GET'])
def oauth_authorized():
    resp = gitlab.authorized_response()
    if resp is None:
        return 'Access denied: reason=%s error=%s' % (
            request.args['error'],
            request.args['error_description']
        )

    token = resp['access_token']
    session['gitlab_token'] = (token, '')
    me = gitlab.get('/api/v4/user')
    user_info = extract_user_info(me.data)
    login = user_login(user_info)
    login.set_cookie('gitlab_token', token)
    return login


@gitlab.tokengetter
def get_gitlab_oauth_token():
    token = session.get('gitlab_token')
    return token
