import logging

from flask import Blueprint
from flask_jwt_extended import jwt_required
from flask_restful import Api, Resource

from portal.rest import course_schema, courses_schema, course_import_schema, rest_helpers
from portal.service.courses import delete_course, update_course, create_course, \
    update_notes_token, copy_course, find_all_courses
from portal.service.filters import filter_course_dump
from portal.service.general import find_course
from portal.service.errors import ForbiddenError
from portal.service.permissions import check_client
from portal.service.auth import find_client
from portal.tools.decorators import error_handler
from portal.service import permissions

courses = Blueprint('courses', __name__)
courses_api = Api(courses)

log = logging.getLogger(__name__)


class CourseResource(Resource):
    @error_handler
    @jwt_required
    def get(self, cid):
        client = find_client()
        course = find_course(cid)
        # authorization
        if (check_client(client=client,
                         course=course,
                         permissions=['view_course_full'])):
            return course_schema.dump(course)

        elif (check_client(client=client,
                           course=course,
                           permissions=['view_course_limited'])):
            dump = course_schema.dump(course)
            filtered_course = filter_course_dump(course, dump.data, client)
            return filtered_course

        raise ForbiddenError(uid=client.id)

    @error_handler
    @jwt_required
    def delete(self, cid):
        client = find_client()
        # authorization
        if not permissions.check_sysadmin(client):
            raise ForbiddenError(uid=client.id)

        course = find_course(cid)
        delete_course(course)
        return '', 204

    @error_handler
    @jwt_required
    def put(self, cid):
        client = find_client()
        course = find_course(cid)
        # authorization
        if not (check_client(client=client, course=course, permissions=['update_course'])):
            raise ForbiddenError(uid=client.id)

        data = rest_helpers.parse_request_data(
            schema=course_schema, action='update', resource='course'
        )
        update_course(course, data)
        return '', 204


class CourseList(Resource):
    @error_handler
    @jwt_required
    def get(self):
        client = find_client()
        # authorization
        if not permissions.check_sysadmin(client):
            raise ForbiddenError(uid=client.id)

        courses_list = find_all_courses()
        return courses_schema.dump(courses_list)

    @error_handler
    @jwt_required
    def post(self):
        client = find_client()
        # authorization
        if not permissions.check_sysadmin(client):
            raise ForbiddenError(uid=client.id)

        data = rest_helpers.parse_request_data(course_schema, resource='course', action='create')
        new_course = create_course(data)
        return course_schema.dump(new_course)[0], 201


class CourseNotesToken(Resource):
    @error_handler
    @jwt_required
    def get(self, cid):
        client = find_client()
        course = find_course(cid)
        # authorization
        if not (check_client(client=client, course=course,
                             permissions=['handle_notes_access_token'])):
            raise ForbiddenError(uid=client.id)

        return course.notes_access_token

    @error_handler
    @jwt_required
    def put(self, cid):
        client = find_client()
        course = find_course(cid)
        # authorization
        if not (check_client(client=client, course=course,
                             permissions=['handle_notes_access_token'])):
            raise ForbiddenError(uid=client.id)

        json_data = rest_helpers.require_data(action='update_notes_token', resource='course')

        update_notes_token(course, json_data)
        return '', 204


class CourseImport(Resource):
    @error_handler
    @jwt_required
    def put(self, cid):
        client = find_client()
        course = find_course(cid)
        # authorization
        if not check_client(client=client, course=course, permissions=['update_course']):
            raise ForbiddenError(uid=client.id)

        data = rest_helpers.parse_request_data(
            course_import_schema, action='import', resource='course'
        )
        source_course = find_course(data['source_course'])
        config = data['config']

        copy_course(source_course, course, config)
        return course_schema.dump(course), 204


courses_api.add_resource(CourseResource, '/<string:cid>')
courses_api.add_resource(CourseList, '')
courses_api.add_resource(CourseNotesToken, "/<string:cid>/notes_access_token")
courses_api.add_resource(CourseImport, "/<string:cid>/import")
