import json
import logging

from flask import Blueprint, request, jsonify
from flask_restful import Api, Resource, abort
from flask_jwt_extended import jwt_required

from portal.rest import project_schema, projects_schema, config_schema, submissions_schema, \
    submission_schema, submission_create_schema, rest_helpers
from portal.service import auth, general
from portal.service.permissions import check_client
from portal.service.projects import delete_project, update_project, create_project, \
    update_project_config, find_project_submissions, can_create_submission, list_projects
from portal.service.submissions import create_submission
from portal.tools.decorators import error_handler
from portal.service.errors import ForbiddenError

projects = Blueprint('projects', __name__)
projects_api = Api(projects)

log = logging.getLogger(__name__)


class ProjectResource(Resource):
    @error_handler
    @jwt_required
    def get(self, cid, pid):
        client = auth.find_client()
        course = general.find_course(cid)
        # authorization
        permissions = ['view_course_full', 'view_course_limited']
        if not (check_client(client=client, course=course, permissions=permissions)):
            raise ForbiddenError(uid=client.id)

        project = general.find_project(course, pid)
        return project_schema.dump(project)

    @error_handler
    @jwt_required
    def delete(self, cid, pid):
        client = auth.find_client()
        course = general.find_course(cid)
        # authorization
        if not check_client(client=client, course=course, permissions=['update_course']):
            raise ForbiddenError(uid=client.id)

        project = general.find_project(course, pid)
        delete_project(project)
        return '', 204

    @error_handler
    @jwt_required
    def put(self, cid, pid):
        client = auth.find_client()
        course = general.find_course(cid)
        # authorization
        permissions = ['write_projects', 'update_course']
        if not check_client(client=client, course=course, permissions=permissions):
            raise ForbiddenError(uid=client.id)

        data = rest_helpers.parse_request_data(
            project_schema, action='update', resource='project'
        )

        project = general.find_project(course, pid)
        update_project(project, data)
        return '', 204


class ProjectsList(Resource):
    @error_handler
    @jwt_required
    def get(self, cid):
        client = auth.find_client()
        course = general.find_course(cid)
        return projects_schema.dump(list_projects(course, client))

    @error_handler
    @jwt_required
    def post(self, cid):
        client = auth.find_client()
        course = general.find_course(cid)
        # authorization
        if not check_client(client=client, course=course, permissions=['update_course']):
            raise ForbiddenError(uid=client.id)

        data = rest_helpers.parse_request_data(
            project_schema, action='create', resource='project'
        )

        new_project = create_project(course, data)
        return project_schema.dump(new_project)[0], 201


class ProjectConfigResource(Resource):
    @error_handler
    @jwt_required
    def get(self, cid, pid):
        client = auth.find_client()
        course = general.find_course(cid)
        # authorization
        permissions = ['read_projects', 'write_projects', 'update_course']
        if not check_client(client=client, course=course, permissions=permissions):
            raise ForbiddenError(uid=client.id)

        project = general.find_project(course, pid)
        return config_schema.dump(project.config)

    @error_handler
    @jwt_required
    def put(self, cid, pid):
        client = auth.find_client()
        course = general.find_course(cid)
        # authorization
        permissions = ['write_projects', 'update_course']
        if not check_client(client=client, course=course, permissions=permissions):
            raise ForbiddenError(uid=client.id)

        data = rest_helpers.parse_request_data(
            config_schema, action='update', resource='project_config'
        )

        project = general.find_project(course, pid)
        update_project_config(project, data)
        return '', 204


class ProjectSubmissions(Resource):
    @error_handler
    @jwt_required
    def get(self, cid, pid):
        client = auth.find_client()
        course = general.find_course(cid)
        # authorization
        if not check_client(client=client, course=course, permissions=['read_submissions_all']):
            raise ForbiddenError(uid=client.id)

        user_id = request.args.get('user')
        project = general.find_project(course, pid)
        submissions = find_project_submissions(project, user_id)
        return submissions_schema.dump(submissions)

    @error_handler
    @jwt_required
    def post(self, cid, pid):
        client = auth.find_client()
        course = general.find_course(cid)
        # authorization
        if not check_client(client=client, course=course, permissions=['create_submissions']):
            raise ForbiddenError(uid=client.id)

        # check if a new submission can be created by this user in this project
        # TODO: configure conflict_timeout in project or
        # remove completely (Kontr handles its own queue)
        project = general.find_project(course, pid)
        if not can_create_submission(client, project):
            # maybe use 409, 403
            abort(429, message=f"New submission for user {client.username} denied: "
                               f"a submission for this user "
                               f"is created or queued.")

        data = rest_helpers.parse_request_data(
            submission_create_schema, action='create', resource='submission'
        )

        # data for Kontr processing, of no use to portal or submission q
        project_params = data.get('project_params')
        project_params_string = json.dumps(project_params) or ""

        new_submission = create_submission(
            user=client,
            project=project,
            file_params=data['file_params'],
            project_params_string=project_params_string
        )
        return submission_schema.dump(new_submission), 201


class ProjectTestFiles(Resource):
    @error_handler
    def get(self, cid, pid):
        # Gets test files (feature)
        return jsonify({"message": "Not implemented."})


projects_api.add_resource(ProjectResource, '/<string:pid>')
projects_api.add_resource(ProjectsList, '')
projects_api.add_resource(ProjectConfigResource, '/<string:pid>/config')
projects_api.add_resource(ProjectSubmissions, '/<string:pid>/submissions')
projects_api.add_resource(ProjectTestFiles, '/<string:pid>/files')
