import logging

from flask import Blueprint
from flask_jwt_extended import jwt_required
from flask_restful import Api, Resource

from portal.rest import role_schema, roles_schema, users_schema, user_list_update_schema, \
    permissions_schema, rest_helpers
from portal.service import auth, general
from portal.service.errors import ForbiddenError
from portal.service.permissions import check_client
from portal.service.roles import delete_role, update_role, create_role, update_role_permissions, \
    update_users_membership_in_role, add_single_user_to_role, remove_single_user_from_role, \
    list_roles
from portal.tools.decorators import error_handler

roles = Blueprint('roles', __name__)
roles_api = Api(roles)

log = logging.getLogger(__name__)


class RoleResource(Resource):
    @error_handler
    @jwt_required
    def get(self, cid, rid):
        client = auth.find_client()
        course = general.find_course(cid)
        # authorization
        permissions = ['view_course_full', 'view_course_limited', 'update_course']
        if not (check_client(client=client, course=course, permissions=permissions)):
            raise ForbiddenError(uid=client.id)

        role = general.find_role(course, rid)
        return role_schema.dump(role)

    @error_handler
    @jwt_required
    def delete(self, cid, rid):
        client = auth.find_client()
        course = general.find_course(cid)
        # authorization
        if not check_client(client=client, course=course, permissions=['update_course']):
            raise ForbiddenError(uid=client.id)

        role = general.find_role(course, rid)
        delete_role(role)
        return '', 204

    @error_handler
    @jwt_required
    def put(self, cid, rid):
        client = auth.find_client()
        course = general.find_course(cid)
        # authorization
        if not check_client(client=client, course=course, permissions=['write_roles']):
            raise ForbiddenError(uid=client.id)

        data = rest_helpers.parse_request_data(
            role_schema, action='update', resource='role'
        )

        role = general.find_role(course, rid)
        update_role(role, data)
        return '', 204


class RoleList(Resource):
    @error_handler
    @jwt_required
    def get(self, cid):
        client = auth.find_client()
        course = general.find_course(cid)
        return roles_schema.dump(list_roles(course, client))

    @error_handler
    @jwt_required
    def post(self, cid):
        client = auth.find_client()
        course = general.find_course(cid)
        # authorization
        if not check_client(client=client, course=course, permissions=['update_course']):
            raise ForbiddenError(uid=client.id)

        data = rest_helpers.parse_request_data(
            role_schema, action='create', resource='role'
        )
        new_role = create_role(course, data)
        return role_schema.dump(new_role)[0], 201


class RolePermissions(Resource):
    @error_handler
    @jwt_required
    def get(self, cid, rid):
        client = auth.find_client()
        course = general.find_course(cid)
        # authorization
        permissions = ['read_roles', 'write_roles', 'update_course']
        if not check_client(client=client, course=course, permissions=permissions):
            raise ForbiddenError(uid=client.id)

        role = general.find_role(course, rid)
        return permissions_schema.dump(role.permissions)

    @error_handler
    @jwt_required
    def put(self, cid, rid):
        client = auth.find_client()
        course = general.find_course(cid)
        # authorization
        permissions = ['write_roles', 'update_course']
        if not check_client(client=client, course=course, permissions=permissions):
            raise ForbiddenError(uid=client.id)

        data = rest_helpers.parse_request_data(
            permissions_schema, action='update', resource='role_permissions'
        )

        role = general.find_role(course, rid)
        update_role_permissions(role, data)
        return permissions_schema.dump(role.permissions), 204


class RoleUsersList(Resource):
    @error_handler
    @jwt_required
    def get(self, cid, rid):
        client = auth.find_client()
        course = general.find_course(cid)
        # authorization
        permissions = ['read_roles', 'write_roles', 'update_course']
        if not check_client(client=client, course=course, permissions=permissions):
            raise ForbiddenError(uid=client.id)

        role = general.find_role(course, rid)
        return users_schema.dump(role.users)

    @error_handler
    @jwt_required
    def put(self, cid, rid):
        client = auth.find_client()
        course = general.find_course(cid)
        # authorization
        permissions = ['write_roles', 'update_course']
        if not check_client(client=client, course=course, permissions=permissions):
            raise ForbiddenError(uid=client.id)

        data = rest_helpers.parse_request_data(
            user_list_update_schema, action='update', resource='role_membership'
        )

        # everything from users_add is added, THEN everything from users_remove is subtracted
        role = general.find_role(course, rid)
        update_users_membership_in_role(role, data)
        return '', 204


class RoleUser(Resource):
    @error_handler
    @jwt_required
    def put(self, cid, rid, uid):
        client = auth.find_client()
        course = general.find_course(cid)

        # authorization
        permissions = ['write_roles', 'update_course']
        if not check_client(client=client, course=course, permissions=permissions):
            raise ForbiddenError(uid=client.id)

        role = general.find_role(course, rid)
        user = general.find_user(uid)
        add_single_user_to_role(role, user)
        return '', 204

    @error_handler
    @jwt_required
    def delete(self, cid, rid, uid):
        client = auth.find_client()
        course = general.find_course(cid)
        # authorization
        permissions = ['write_roles', 'update_course']
        if not check_client(client=client, course=course, permissions=permissions):
            raise ForbiddenError(uid=client.id)

        role = general.find_role(course, rid)
        user = general.find_user(uid)
        remove_single_user_from_role(role, user)
        return '', 204


roles_api.add_resource(RoleResource, "/<string:rid>")
roles_api.add_resource(RoleList, "")
roles_api.add_resource(RolePermissions, "/<string:rid>/permissions")
roles_api.add_resource(RoleUsersList, "/<string:rid>/users")
roles_api.add_resource(RoleUser, "/<string:rid>/users/<string:uid>")
