from flask_jwt_extended import get_jwt_identity, get_jwt_claims
from typing import Union

from portal.database.models import User, Component
from portal.service.errors import UnauthorizedError, PortalAPIError, IncorrectCredentialsError
from portal.service.general import find_component, find_user


def login_user(gitlab_access_token: str = None, password: str = None, username: str = None) -> User:
    """Verifies user login credentials. Valid parameter combinations are:
     - username, password,
     - username, gitlab_access_token

    Args:
        gitlab_access_token(str): OAuth access token received from GitLab
        password(str): user's password
        username(str): username to log in

    Returns(User): the authenticated user

    """
    if gitlab_access_token:
        return auth_gitlab_access_token(
            username=username,
            gitlab_access_token=gitlab_access_token
        )

    return auth_username_password(username=username, password=password)


def auth_gitlab_access_token(username: str, gitlab_access_token: str) -> User:
    """Verifies GitLab access token and username combination.

    Args:
        username(str): username of the user attempting to log in
        gitlab_access_token(str): access token string from gitlab

    Returns(User): the authenticated user
    """
    if gitlab_access_token is None:
        raise PortalAPIError(400, 'No gitlab access token found.')
    # TODO: validate gitlab token
    user = find_user(username, throws=False)
    if user is None:
        raise IncorrectCredentialsError()
    return user


def auth_username_password(username: str, password: str) -> User:
    """Authorizes the user using username and password
    Args:
        username(str): Username
        password(str): Password
    Returns(User): the authenticated user
    """
    user = find_user(username, throws=False)
    if user is None or password is None:
        raise IncorrectCredentialsError()

    if user.verify_password(password=password):
        return user

    raise IncorrectCredentialsError()


def login_component(name: str, ip_address: str, secret: str) -> Component:
    """Login a component of the system
    Args:
        name(str): Name of the component
        ip_address(str): IP address of the component
        secret(str): Secret code of the component
    Returns(Component): the authenticated component
    """
    client = find_component(name)
    if client.secret != secret:
        raise UnauthorizedError(note="Invalid component secret.")
    if ip_address != client.ip_address:
        raise UnauthorizedError(note="Different source ip address for component login request.")
    return client


def find_client() -> Union[User, Component]:
    """Gets a user/component based on the jwt token

    Returns(User, Component): User or Component instance
    """
    identity = get_jwt_identity()
    claims = get_jwt_claims()
    if not (claims or claims.get('type')):
        raise UnauthorizedError(note="No type claim.")
    if claims['type'] == 'user':
        client = __find_client_helper(identity, 'user', find_user)
    elif claims['type'] == 'component':
        client = __find_client_helper(identity, 'component', find_component)
    else:
        raise UnauthorizedError(note="Invalid jwt claims.")
    return client


def __find_client_helper(identity, resource, method):
    client = method(identity, throws=False)
    if not client:
        raise UnauthorizedError(note=f"{resource.upper()} with identity {identity} not found.")
    return client
