import logging

from portal.database.models import Course, Role, RolePermissions, User
from portal.service import general, filters
from portal.service.errors import PortalAPIError, ForbiddenError
from portal.service.general import find_role, delete_entity, write_entity
from portal.service.permissions import check_client
from portal.service.users import find_users

log = logging.getLogger(__name__)


def copy_role(source: Role, target: Course, with_users: str) -> Role:
    """Copies role to the target course
    Args:
        source(Role): Source role resource
        target(Course): Target Course resource
        with_users(str): With users flag

    Returns(Role): Copied role
    """
    log.info(f"[COPY] Role: {source.id} to course {target.id} with config: {with_users}")
    new_name = source.name
    count = 1
    if find_role(target, source.name, throws=False):
        new_name += "_copy0"
    while find_role(target, new_name, throws=False):
        new_name[-1] = count
        count += 1
    new_role = Role(target, new_name)
    new_role.set_permissions(**vars(source.permissions))
    new_role.description = source.description
    if with_users == "with_users":
        for user in source.users:
            user.roles.append(new_role)
    return new_role


def delete_role(role: Role):
    """Deletes role for the DB
    Args:
        role(Role): Role instance
    """
    log.info(f"[DELETE] Role: {role.id} ({role.name})")
    delete_entity(role)


def update_role(role: Role, data: dict) -> Role:
    """Updates role based on data
    Args:
        role(Role): Role instance
        data(dict): Data for role update
    Returns(Role): Updated role
    """
    log.info(f"[UPDATE] Role {role.id} to {role}.")
    return __set_role_data(role, data)


def __set_role_data(role: Role, data: dict):
    return general.update_entity(role, data, allowed=['name', 'description'])


def create_role(course: Course, data: dict) -> Role:
    """Creates new role
    Args:
        course(Course): Course instance
        data(dict): Role data
    Returns(Role): New role instance
    """
    new_role = Role(course=course)
    __set_role_data(role=new_role, data=data)
    log.info(f"[CREATE] Role for course {course.id}: {new_role}")
    return new_role


def update_role_permissions(role: Role, permissions: dict) -> RolePermissions:
    """Updates role permissions
    Args:
        role(Role): Role instance
        permissions(dict): Permissions dictionary

    Returns(RolePermissions): Role Permissions instance
    """
    role.set_permissions(**permissions)
    log.info(f"[UPDATE] Permissions for role {role.id} "
             f"in course {role.course.id}: {permissions}.")
    return role.permissions


def update_users_membership_in_role(role: Role, data: dict) -> Role:
    """Updates users membership in the role
    Args:
        role(Role): Role instance
        data(dict): Data provided to update role membership

    Returns(Role): Updated role
    """
    users_old = set(role.users)
    users_add = data.get('add')
    users_remove = data.get('remove')
    users = set()
    if users_add:
        to_add = find_users(data['add'])
        users = users_old.union(to_add)
    if users_remove:
        to_remove = find_users(data['remove'])
        users = users_old.difference(to_remove)
    role.users = list(users)
    write_entity(role)
    log.info(f"[UPDATE] Users in Role {role.id}: {data}.")
    return role


def add_single_user_to_role(role: Role, user: User):
    """Adds single user to the role
    Args:
        role(Role): Role instance
        user(User): User instance
    Returns:
    """
    if user not in role.users:
        role.users.append(user)
        log.info(f"[ADD] User {user.id} to role "
                 f"{role.id} in course {role.course.id}.")
    else:
        log.info(f"[ADD] User {user.id} is already "
                 f"in role {role.id} in course {role.course.id}: no change.")


def remove_single_user_from_role(role: Role, user: User) -> Role:
    """Removes single user from the role
    Args:
        role(Role): Role instance
        user(User): User instance
    Returns(Role): Updated role
    """
    course = role.course
    try:
        role.users.remove(user)
    except ValueError:
        raise PortalAPIError(400, message=f"Could not remove user {user.id} "
                                          f"from role {role.id} in course {course.id}: "
                                          f"role does not contain user.")
    log.info(f"[REMOVE] User {user.id} from role {role.id} in course {course.id}.")
    return role


def list_roles(course, client):
    if check_client(client=client, course=course, permissions=['view_course_full']):
        return course.roles
    elif check_client(client=client, course=course, permissions=['view_course_limited']):
        return filters.filter_roles_from_course(course=course, user=client)
    raise ForbiddenError(uid=client.id)
